#include "cmdline_maxlink-query1.h"

#include <cstdlib>
#include <stdint.h>
#include <iostream>
#include <algorithm>
#include <set>
#include <boost/foreach.hpp>
#include <boost/test/utils/foreach.hpp>
#include <boost/math/distributions/hypergeometric.hpp>
#include <boost/math/policies/policy.hpp>

#include "fbs_database.h"
#include "edge_database.h"
#include "num2string_gene_database.h"
#include "string2num_gene_database.h"
#include "maxlink_xml_parser.h"
#include "maxlink_text_parser.h"
#include "maxlink_gene_list_xml_parser.h"
#include "maxlink_types.h"
#include "config.h"
#include "common_funcs.h"

std::vector<maxlink_types::GeneId> geneListFromUserInput(const struct gengetopt_args_info &args_info, String2NumGeneDatabase *string2num_gene_database) {
  std::vector<maxlink_types::GeneId> gene_list;
  switch (args_info.gene_list_file_format_arg) {
    case gene_list_file_format_arg_xml: {
      MaxlinkGeneListXmlParser maxlink_gene_list_xml_parser;
      return maxlink_gene_list_xml_parser.loadGeneVectorFromXml(args_info.gene_list_filename_arg, string2num_gene_database);
      break;
    }
    case gene_list_file_format_arg_text: {
      std::string line;
      std::ifstream file(args_info.gene_list_filename_arg);
      maxlink_types::GeneId gene;
      while(std::getline(file, line)) {
        if (!string2num_gene_database->translateString2Num(line, &gene)) {
          std::cerr << "error: could not translate gene id string \"" << line << "\" to gene id number" << std::endl;     
          // exit(EXIT_FAILURE);
        } else {
          gene_list.push_back(gene);
        }
      }
      break;
    }
    default: assert(false); /* indicate programming error */ 
  }
  return gene_list;
}

static bool sortGeneIdScore(maxlink_types::GeneIdScore u,  maxlink_types::GeneIdScore v)
{
	if (v.maxlink_score == u.maxlink_score)
		return u.non_q_neighbours_count < v.non_q_neighbours_count;
	return v.maxlink_score < u.maxlink_score;
}

static bool sortHypergeometric(maxlink_types::GeneIdScore u,  maxlink_types::GeneIdScore v)
{
	if (v.maxlink_score == u.maxlink_score)
		return u.hypergeometric_probability < v.hypergeometric_probability;
	return v.maxlink_score < u.maxlink_score;
}

std::vector<maxlink_types::GeneIdScore> calculateMaxlinkVersion1(const struct gengetopt_args_info &args_info, EdgeDatabase *edge_database, FbsDatabase *fbs_database, std::vector<maxlink_types::GeneId> &q_vector, maxlink_types::GeneCount *q_edges_total_count_ptr, maxlink_types::GeneCount *non_q_edges_total_count_ptr, maxlink_types::GeneCount *non_q_genes_with_edge_to_q_count_ptr) {
  std::sort(q_vector.begin(), q_vector.end());  

  if (args_info.debug_flag) {  
    std::cerr << "debug: in function calculateMaxlinkVersion1()" << std::endl;
    std::cerr << "debug: edge_database->num_entries() = " << edge_database->num_entries() << std::endl;
    std::cerr << "debug: fbs_database->num_entries() = " << fbs_database->num_entries() << std::endl;
    std::cerr << "debug: q_vector.size() = " << q_vector.size() << std::endl;
  }
  std::set<maxlink_types::GeneId> set_of_non_q_genes_with_edge_to_q;

  //  q_edges_total_count is the number of links, where at least one of the genes in the pair is from the Q set.
  maxlink_types::GeneCount q_edges_total_count = 0; 
  BOOST_FOREACH(const maxlink_types::GeneId it, q_vector) {
    std::vector<maxlink_types::GeneId> neighbours = edge_database->neighbours(args_info.debug_flag, it, args_info.fbs_cutoff_arg);   
    if (args_info.debug_flag) {  
      std::cerr << "debug: neighbours.size() = " << neighbours.size() << std::endl;
    }
    std::sort(neighbours.begin(), neighbours.end());
    std::vector<maxlink_types::GeneId> neighbours_that_are_nonQ;
    std::set_difference(neighbours.begin(), neighbours.end(), q_vector.begin(), q_vector.end(), std::inserter(neighbours_that_are_nonQ, neighbours_that_are_nonQ.end()));
    set_of_non_q_genes_with_edge_to_q.insert(neighbours_that_are_nonQ.begin(), neighbours_that_are_nonQ.end());
    // An edge to a Q-gene is only given half of a normal count (+2) because later the same edge will be found again and thus counted twice. After counting we divide by 2 to get the correct count.
    q_edges_total_count += (2 * neighbours_that_are_nonQ.size() + neighbours.size() - neighbours_that_are_nonQ.size()); 
  }
  assert(q_edges_total_count % 2 == 0);
  q_edges_total_count = q_edges_total_count / 2;
  if (args_info.debug_flag) {  
    std::cerr << "debug: fbs_database->num_edges_over_cutoff(args_info.fbs_cutoff_arg) = " << fbs_database->num_edges_over_cutoff(args_info.fbs_cutoff_arg) << std::endl;
    std::cerr << "debug: q_edges_total_count = " <<  q_edges_total_count << std::endl;
  }
  maxlink_types::GeneCount non_q_edges_total_count = fbs_database->num_edges_over_cutoff(args_info.fbs_cutoff_arg) - q_edges_total_count;
  *non_q_edges_total_count_ptr = non_q_edges_total_count;
  double connectivity_cutoff =   non_q_edges_total_count / static_cast<double>(q_edges_total_count);
  *q_edges_total_count_ptr = q_edges_total_count;
  std::vector<maxlink_types::GeneIdScore> result_vec;
  *non_q_genes_with_edge_to_q_count_ptr =  set_of_non_q_genes_with_edge_to_q.size();

  BOOST_FOREACH(maxlink_types::GeneId it, set_of_non_q_genes_with_edge_to_q) {
    std::vector<maxlink_types::GeneId> neighbours2 = edge_database->neighbours(args_info.debug_flag, it, args_info.fbs_cutoff_arg);   
    std::sort(neighbours2.begin(), neighbours2.end());  
    std::vector<maxlink_types::GeneId> non_q_neighbours_vector;
    std::set_difference(neighbours2.begin(), neighbours2.end(), q_vector.begin(), q_vector.end(), std::inserter(non_q_neighbours_vector, non_q_neighbours_vector.end()));
    maxlink_types::GeneIdScore gene_id_score;
    gene_id_score.gene_id = it;
    gene_id_score.maxlink_score =  neighbours2.size()-non_q_neighbours_vector.size();
    gene_id_score.non_q_neighbours_count = non_q_neighbours_vector.size();  // $nonQorKlinks

    // Read more about gsl_cdf_hypergeometric_Q() here:
    // http://www.gnu.org/software/gsl/manual/html_node/The-Hypergeometric-Distribution.html
    unsigned int k = gene_id_score.maxlink_score;
    unsigned int n1 = q_edges_total_count;
    unsigned int n2 = non_q_edges_total_count;
    unsigned int t = gene_id_score.non_q_neighbours_count + gene_id_score.maxlink_score;

    // The GNU Scientific Library functions were replaced with Boost functions. For more info see:
    // http://stackoverflow.com/questions/13274665/translate-gsl-ran-hypergeometric-pdfk-n1-n2-t-to-boostmathhypergeometri/13274706#13274706
    // gene_id_score.hypergeometric_probability = gsl_ran_hypergeometric_pdf(k, n1, n2, t) + gsl_cdf_hypergeometric_Q(k, n1, n2, t);
    boost::math::hypergeometric_distribution<double> hg_dist(t, n1, n1 + n2);
    gene_id_score.hypergeometric_probability = boost::math::pdf<double>(hg_dist, k) + (1 - boost::math::cdf<double>(hg_dist, k));
    if (args_info.debug_flag) {  
      std::cerr << "debug: calculating hypergeometric_probability" << std::endl;
      std::cerr << "debug: k = " << k << std::endl;
      std::cerr << "debug: n1 = " << n1 << std::endl;
      std::cerr << "debug: n2 = " << n2 << std::endl;
      std::cerr << "debug: t = " << t << std::endl;
      std::cerr << "debug: gene_id_score.hypergeometric_probability = " << gene_id_score.hypergeometric_probability << std::endl;  
  }
    switch (args_info.result_method_arg) {
      case result_method_arg_ratio : { 
        if (static_cast<double>(gene_id_score.non_q_neighbours_count) / gene_id_score.maxlink_score < connectivity_cutoff) {
          result_vec.push_back(gene_id_score);
        }
        break; 
      };
      case result_method_arg_hypergeometric : {
        if (args_info.probability_cutoff_arg >= gene_id_score.hypergeometric_probability) {
          result_vec.push_back(gene_id_score);
        }
        break;
      };
      default: assert(false); // indicate programming error
    }
  }
  switch (args_info.result_method_arg) {
    case result_method_arg_ratio : { 
      std::sort(result_vec.begin(), result_vec.end(), sortGeneIdScore);
      break; 
    };
    case result_method_arg_hypergeometric : {
      std::sort(result_vec.begin(), result_vec.end(), sortHypergeometric);
      break;
    };
    default: assert(false); // indicate programming error
  }
  return result_vec;
}

int main(int argc, char *argv[]) {
  EdgeDatabase edge_database;
  FbsDatabase fbs_database;
  Num2StringGeneDatabase num2string_gene_database;
  String2NumGeneDatabase string2num_gene_database;
  struct gengetopt_args_info args_info;
  if(cmdline_parser(argc, argv, &args_info) != 0) {
    exit(EXIT_FAILURE);
  }
  char *tokyo_cabinet_dir = get_tokyo_cabinet_dir(args_info.tokyo_cabinet_dir_given, args_info.tokyo_cabinet_dir_arg);
  assert(tokyo_cabinet_dir);

  std::string fbs_cutoff_file = std::string(tokyo_cabinet_dir) + std::string("/fbs_cutoff");
  std::ifstream cutoff_istream;
  cutoff_istream.open(fbs_cutoff_file.c_str()); 
  if (!cutoff_istream.good()) {
    std::cerr << "error: Could not open the file fbs_cutoff in the tokyo cabinet dir" << std::endl;
    exit(EXIT_FAILURE);
  }
  float fbs_cutoff_used_in_maxlink_load;
  cutoff_istream >> fbs_cutoff_used_in_maxlink_load;

  if (args_info.debug_flag) {  
    std::cerr << "debug: from file \"" << fbs_cutoff_file << "\" the fbs_cutoff " << fbs_cutoff_used_in_maxlink_load << " was read" << std::endl;
  }

  if (args_info.fbs_cutoff_arg < fbs_cutoff_used_in_maxlink_load) {
    std::cerr << "error: --fbs-cutoff value is smaller than the --fbs-cutoff value that was used in maxlink-load" << std::endl;
    exit(EXIT_FAILURE);
  }
  num2string_gene_database.openReadOnly(tokyo_cabinet_dir);
  string2num_gene_database.openReadOnly(tokyo_cabinet_dir);
  edge_database.openReadOnly(tokyo_cabinet_dir);
  fbs_database.openReadOnly(tokyo_cabinet_dir);
  std::vector<maxlink_types::GeneId> q_vector = geneListFromUserInput(args_info, &string2num_gene_database);
  maxlink_types::GeneCount q_edges_total_count;
  maxlink_types::GeneCount non_q_edges_total_count;
  maxlink_types::GeneCount non_q_genes_with_edge_to_q_count;
  std::vector<maxlink_types::GeneIdScore> maxlink_result1 = calculateMaxlinkVersion1(args_info, &edge_database, &fbs_database, q_vector, &q_edges_total_count, &non_q_edges_total_count, &non_q_genes_with_edge_to_q_count);
  switch (args_info.output_format_arg) {
    case output_format_arg_tsv: {
      std::cout << "#Gene ID\tMaxLink score\tP-value\tRatio" << std::endl;
      BOOST_FOREACH(maxlink_types::GeneIdScore it, maxlink_result1) {
        // The Coogle C++ Style Guide recommends fprintf instead of iostream     
        // http://google-styleguide.googlecode.com/svn/trunk/cppguide.xml?showone=Streams#Streams
        // But if we use printf we also need to take care of 64-bit portability:
        // http://google-styleguide.googlecode.com/svn/trunk/cppguide.xml?showone=64-bit_Portability#64-bit_Portability
        // 
        // Anyway just for now we use iostream
        std::cout << num2string_gene_database.translateNum2String(it.gene_id) << "\t" << it.maxlink_score  << "\t" << it.hypergeometric_probability <<  "\t" << (static_cast<double>(it.non_q_neighbours_count) / it.maxlink_score) << std::endl;
      }
      break;   
    }  
    case output_format_arg_xml: {
      std::cout << "<?xml version=\"1.0\"?>" << std::endl  << "<maxlink_query1_result xmlns=\"" << MAXLINK_QUERY1_RESULT_NAMESPACE << "\" q_edges_total_count=\"" << q_edges_total_count
  << "\" non_q_genes_with_edge_to_q_count=\"" << non_q_genes_with_edge_to_q_count 
                << "\" " << std::endl << "non_q_edges_total_count=\"" << non_q_edges_total_count << "\" connectivity_cutoff=\"" <<  non_q_edges_total_count / static_cast<double>(q_edges_total_count)  << "\">" << std::endl;   
      BOOST_FOREACH(maxlink_types::GeneIdScore it, maxlink_result1) {
        // The Coogle C++ Style Guide recommends fprintf instead of iostream     
        // http://google-styleguide.googlecode.com/svn/trunk/cppguide.xml?showone=Streams#Streams
        // But if we use printf we also need to take care of 64-bit portability:
        // http://google-styleguide.googlecode.com/svn/trunk/cppguide.xml?showone=64-bit_Portability#64-bit_Portability
        // 
        // Anyway just for now we use iostream
        std::cout << "<gene gene_id=\"" << num2string_gene_database.translateNum2String(it.gene_id) 
                  << "\" score=\"" << it.maxlink_score  << "\" hypergeometric_probability=\"" << it.hypergeometric_probability << "\" non_q_neighbours_count=\"" << it.non_q_neighbours_count << "\"/>"  << std::endl;   
      }
      std::cout << "</maxlink_query1_result>" << std::endl;   
      break;   
    }
    default: assert(false); /* indicate programming error */ 
  }
}
